searchNodes=[{"doc":"Google's XLA (Accelerated Linear Algebra) compiler/backend for Nx. It supports just-in-time (JIT) compilation to GPU (both CUDA and ROCm) and TPUs. Configuration In projects EXLA works both as a backend for Nx tensors and an optimized Nx.Defn compiler. To enable both globally, add a config/config.exs (or config/ENV.exs ) with the following: import Config config :nx , :default_backend , EXLA.Backend config :nx , :default_defn_options , [ compiler : EXLA ] Now you can use Nx as usual and it will use EXLA by default. You can also use cuda/rocm/tpu as the target by setting :client option in both configuration: import Config config :nx , :default_backend , { EXLA.Backend , client : :cuda } config :nx , :default_defn_options , [ compiler : EXLA , client : :cuda ] To use GPUs/TPUs, you must also set the appropriate value for the XLA_TARGET environment variable. For CUDA, setting ELIXIR_ERL_OPTIONS=&quot;+sssdio 128&quot; is also required on more complex operations to increase CUDA's compiler stack size. In scripts/notebooks The simplest way to configure EXLA in notebooks is by calling: EXLA . set_as_nx_default ( [ :tpu , :cuda , :rocm , :host ] ) Then EXLA will pick the first platform available for the current notebook user. From them on, you can use Nx as usual and it will use EXLA by default. As in the project configuration above, you must also set the appropriate value for the XLA_TARGET environment variable if you intend to use GPU/TPUs. Options The options accepted by EXLA configuration are: :client - an atom representing the client to use. Defaults to :host . See &quot;Clients&quot; section :device_id - the default device id to run the computation on. Defaults to the :default_device_id on the client Clients The EXLA library uses a client for compiling and executing code. Those clients are typically bound to a platform, such as CPU or GPU. Those clients are singleton resources on Google's XLA library, therefore they are treated as a singleton resource on this library too. EXLA ships with the client configuration for each supported platform, which would be the equivalent to this: config :exla , :clients , host : [ platform : :host ] , cuda : [ platform : :cuda ] , rocm : [ platform : :rocm ] , tpu : [ platform : :tpu ] Important! you should avoid using multiple clients for the same platform. If you have multiple clients per platform, they can race each other and fight for resources, such as memory. Therefore, we recommend developers to stick with the default clients above. Client options Each client configuration accepts the following options: :platform - the platform the client runs on. It can be :host (CPU), :cuda , :rocm , or :tpu . :default_device_id - the default device ID to run on. For example, if you have two GPUs, you can choose a different one as the default. Defaults to device 0 (the first device). :preallocate - if the memory should be preallocated on GPU devices. Defaults to true . :memory_fraction - how much memory of a GPU device to allocate. Defaults to 0.9 . GPU Runtime Issues GPU Executions run in dirty IO threads, which have a considerable smaller stack size than regular scheduler threads. This may lead to problems with certain CUDA or cuDNN versions, leading to segmentation fails. In a development environment, it is suggested to set: ELIXIR_ERL_OPTIONS = &quot;+sssdio 128&quot; To increase the stack size of dirty IO threads from 40 kilowords to 128 kilowords. In a release, you can set this flag in your vm.args . Device allocation EXLA also ships with a EXLA.Backend that allows data to be explicitly allocated on the EXLA device. You can create tensors with EXLA.Backend directly: Nx . tensor ( [ 1 , 2 , 3 , 4 ] , backend : EXLA.Backend ) or you can configure EXLA.Backend as the default backend, so that all tensors are allocated on the EXLA device by default. In some cases you may want to explicitly move an existing tensor to the device: tensor = Nx . tensor ( [ 1 , 2 , 3 , 4 ] , backend : Nx.BinaryBackend ) Nx . backend_transfer ( tensor , EXLA.Backend ) Note that you can use regular Nx operations, so the following works: tensor = Nx . tensor ( [ 1 , 2 , 3 , 4 ] , backend : EXLA.Backend ) Nx . sum ( tensor ) Under the hood, EXLA will create a computation for the sum operation and invoke it on the device. This is essentially an &quot;eager mode&quot; that provides acceleration during prototyping. However, eventually you should wrap your computations in a defn to utilize the full performance of JIT. Docker considerations EXLA should run fine on Docker with one important consideration: you must not start the Erlang VM as the root process in Docker. That's because when the Erlang VM runs as root, it has to manage all child programs. At the same time, Google XLA's shells out to child program during compilation and it must retain control over how child programs terminate. To address this, simply make sure you wrap the Erlang VM in another process, such as the shell one. In other words, if you are using releases, instead of this: RUN path / to / release start do this: RUN sh - c &quot;path/to/release start&quot; If you are using Mix inside your Docker containers, instead of this: RUN mix run do this: RUN sh - c &quot;mix run&quot; Alternatively, you can pass the --init flag to docker run , so it runs an init inside the container that forwards signals and reaps processes.","ref":"EXLA.html","title":"EXLA","type":"module"},{"doc":"A shortcut for Nx.Defn.jit/3 with the EXLA compiler. iex&gt; EXLA . jit ( &amp; Nx . add ( &amp;1 , &amp;1 ) , [ Nx . tensor ( [ 1 , 2 , 3 ] ) ] ) # Nx.Tensor &lt; s64 [ 3 ] [ 2 , 4 , 6 ] &gt; See the moduledoc for options.","ref":"EXLA.html#jit/3","title":"EXLA.jit/3","type":"function"},{"doc":"Checks if the JIT compilation of function with args is cached. Note that hooks are part of the cache, and therefore they must be included in the options. Examples iex&gt; fun = fn a , b -&gt; Nx . add ( a , b ) end iex&gt; left = Nx . tensor ( 1 , type : { :u , 8 } ) iex&gt; right = Nx . tensor ( [ 1 , 2 , 3 ] , type : { :u , 16 } ) iex&gt; EXLA . jit ( fun , [ left , right ] ) iex&gt; EXLA . jit_cached? ( fun , [ left , right ] ) true iex&gt; EXLA . jit_cached? ( fun , [ left , Nx . tensor ( [ 1 , 2 , 3 , 4 ] , type : { :u , 16 } ) ] ) false","ref":"EXLA.html#jit_cached?/3","title":"EXLA.jit_cached?/3","type":"function"},{"doc":"Sets the global defn options to the EXLA compiler with the preferred client based on their availability. This function is typically invoked at the top of scripts and code notebooks which might be potentially executed from multiple platforms. Do not invoke this function during runtime, as it changes Nx.Defn options globally. If you have a specific client that you want to use throughout your project, use configuration files instead: import Config config :nx , :default_backend , { EXLA.Backend , client : :cuda } config :nx , :default_defn_options , [ compiler : EXLA , client : :cuda ] Examples EXLA . set_as_nx_default ( [ :tpu , :cuda , :rocm , :host ] ) The above will try to find the first client available and set the EXLA compiler with the client as the compilers for Nx.Defn . If no client is found, EXLA is not set as compiler at all, therefore it is common to add :host as the last option. If additional options are given, they are given as compiler options: EXLA . set_as_nx_default ( [ :tpu , :cuda , :rocm , :host ] ) To use the GPU or TPUs, don't forget to also set the appropriate value for the XLA_TARGET environment variable.","ref":"EXLA.html#set_as_nx_default/2","title":"EXLA.set_as_nx_default/2","type":"function"},{"doc":"Starts streaming the given anonymous function with just-in-time compilation. At least two arguments are expected: The first argument is a tensor template of the data to be streamed in The second argument is a tensor with the stream initial state The streaming function must return a two element tuple, the first element is the data to be sent and the second is the accumulator. For each streamed chunk, you must call Nx.Stream.send/2 and Nx.Stream.recv/1 . You don't need to call recv immediately after send , but doing so can be a useful mechanism to provide backpressure. Once all chunks are sent, you must use Nx.Stream.done/1 to receive the accumulated result. Let's see an example: defmodule Streamed do import Nx.Defn defn sum ( tensor , acc ) do { acc , tensor + acc } end end Now let's invoke it: stream = EXLA . stream ( &amp; Streamed . sum / 2 , [ Nx . template ( { } , { :s , 64 } ) , 0 ] ) for i &lt;- 1 .. 5 do Nx.Stream . send ( stream , i ) IO . inspect { :chunk , Nx.Stream . recv ( stream ) } end IO . inspect { :result , Nx.Stream . done ( stream ) } It will print: { :chunk , 0 } { :chunk , 1 } { :chunk , 2 } { :chunk , 3 } { :chunk , 4 } { :result , 5 } Note: While any process can call Nx.Stream.send/2 , EXLA expects the process that starts the streaming to be the one calling Nx.Stream.recv/1 and Nx.Stream.done/1 .","ref":"EXLA.html#stream/3","title":"EXLA.stream/3","type":"function"},{"doc":"Checks if the JIT compilation of stream with args is cached. Note that hooks are part of the cache, and therefore they must be included in the options. Examples iex&gt; left = Nx . tensor ( 1 , type : { :u , 8 } ) iex&gt; right = Nx . tensor ( [ 1 , 2 , 3 ] , type : { :u , 16 } ) iex&gt; fun = fn x , acc -&gt; { acc , Nx . add ( x , acc ) } end iex&gt; stream = EXLA . stream ( fun , [ left , right ] ) iex&gt; Nx.Stream . done ( stream ) iex&gt; EXLA . stream_cached? ( fun , [ left , right ] ) true iex&gt; EXLA . stream_cached? ( fun , [ left , Nx . tensor ( [ 1 , 2 , 3 , 4 ] , type : { :u , 16 } ) ] ) false","ref":"EXLA.html#stream_cached?/3","title":"EXLA.stream_cached?/3","type":"function"},{"doc":"A Nx tensor backend for the data kept on the device. You can directly transfer to this backend by calling Nx.backend_transfer/2 or Nx.backend_copy/2 . It allows the following options: :client - the client to store the data on. Defaults to the client configured in Nx.Defn , otherwise uses :host . :device_id - which device to store it on To get the data out of the device backend into a regular tensor, call Nx.backend_transfer/1 (with the device tensor as the single argument).","ref":"EXLA.Backend.html","title":"EXLA.Backend","type":"module"},{"doc":"A buffer where data is kept in a binary.","ref":"EXLA.BinaryBuffer.html","title":"EXLA.BinaryBuffer","type":"module"},{"doc":"","ref":"EXLA.BinaryBuffer.html#from_binary/2","title":"EXLA.BinaryBuffer.from_binary/2","type":"function"},{"doc":"Wrapper around XLA's builder.","ref":"EXLA.Builder.html","title":"EXLA.Builder","type":"module"},{"doc":"","ref":"EXLA.Builder.html#build/1","title":"EXLA.Builder.build/1","type":"function"},{"doc":"","ref":"EXLA.Builder.html#new/1","title":"EXLA.Builder.new/1","type":"function"},{"doc":"","ref":"EXLA.Builder.html#new/2","title":"EXLA.Builder.new/2","type":"function"},{"doc":"Functions for managing EXLA.Client . See EXLA module docs for a general introduction.","ref":"EXLA.Client.html","title":"EXLA.Client","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"EXLA.Client.html#child_spec/1","title":"EXLA.Client.child_spec/1","type":"function"},{"doc":"Copies buffer to device with given device ID.","ref":"EXLA.Client.html#copy_buffer_to_device/3","title":"EXLA.Client.copy_buffer_to_device/3","type":"function"},{"doc":"Fetches a client with the given name from configuration.","ref":"EXLA.Client.html#fetch!/1","title":"EXLA.Client.fetch!/1","type":"function"},{"doc":"Sends buffer from device outfeed to the given process tagged by ref . Note: XLA does not support tuple outfeed shapes. Passing one will simply block the operation indefinitely. Instead, convert the tuple into multiple outfeed operations.","ref":"EXLA.Client.html#from_outfeed/5","title":"EXLA.Client.from_outfeed/5","type":"function"},{"doc":"Returns a map of supported platforms with device information.","ref":"EXLA.Client.html#get_supported_platforms/0","title":"EXLA.Client.get_supported_platforms/0","type":"function"},{"doc":"Sends data_and_shapes to device infeed. data_and_shapes must be a list of two element tuples where the first element is a binary or a flat list of binaries and the second element is a EXLA.Shape . Note: XLA does not support tuple infeed shapes when running on host. Passing one will simply block the operation indefinitely. Instead, convert the tuple into multiple infeed operations.","ref":"EXLA.Client.html#to_infeed/3","title":"EXLA.Client.to_infeed/3","type":"function"},{"doc":"Wrapper around XLA's computation.","ref":"EXLA.Computation.html","title":"EXLA.Computation","type":"module"},{"doc":"Compiles a computation into an executable. Options :device_id - the device id to compile to and run the executable on. Defaults to the :default_device_id on the client. If :num_replicas or :num_partitions are given, this option is ignored and the device id is set to -1 . :num_replicas - the number of replicas this computation will run on. It defaults to 1 but you can set it if you want to enable single-program multiple data. :use_spmd - enables Single-Program Multiple-Data partioning. This is set to true if :num_partitions is more than one, otherwise is false . Currently those options do not have an effect as they related to running the same compiled executable on multiple replicas. Some options apply to TPU only: :num_partitions - the number of partitions this computation will run on.","ref":"EXLA.Computation.html#compile/4","title":"EXLA.Computation.compile/4","type":"function"},{"doc":"An EXLA DeviceBuffer for data allocated in the device.","ref":"EXLA.DeviceBuffer.html","title":"EXLA.DeviceBuffer","type":"module"},{"doc":"Deallocates the underlying buffer. Returns :ok | :already_deallocated .","ref":"EXLA.DeviceBuffer.html#deallocate/1","title":"EXLA.DeviceBuffer.deallocate/1","type":"function"},{"doc":"Places the given binary buffer on the given device using client .","ref":"EXLA.DeviceBuffer.html#place_on_device/4","title":"EXLA.DeviceBuffer.place_on_device/4","type":"function"},{"doc":"Reads size from the underlying buffer ref. This copies the underlying device memory into a binary without destroying it. If size is negative, then it reads the whole buffer.","ref":"EXLA.DeviceBuffer.html#read/2","title":"EXLA.DeviceBuffer.read/2","type":"function"},{"doc":"Wrapper around XLA's executable.","ref":"EXLA.Executable.html","title":"EXLA.Executable","type":"module"},{"doc":"Runs the given executable with a list of lists as inputs and the given options.","ref":"EXLA.Executable.html#run/3","title":"EXLA.Executable.run/3","type":"function"},{"doc":"High-level operations built on top of EXLA.Op .","ref":"EXLA.Lib.html","title":"EXLA.Lib","type":"module"},{"doc":"Computes the argmax of the given operation. Options :axis - the axis to reduce on :keep_axis - whether or not to keep reduced axis :tie_break - how to break ties","ref":"EXLA.Lib.html#argmax/3","title":"EXLA.Lib.argmax/3","type":"function"},{"doc":"Computes the argmin of the given operation. Options :axis - the axis to reduce on :keep_axis - whether or not to keep reduced axis :tie_break - how to break ties","ref":"EXLA.Lib.html#argmin/3","title":"EXLA.Lib.argmin/3","type":"function"},{"doc":"Sorts a tensor and returns the corresponding indices in the new positions.","ref":"EXLA.Lib.html#argsort/5","title":"EXLA.Lib.argsort/5","type":"function"},{"doc":"Builds iota along axis.","ref":"EXLA.Lib.html#iota/3","title":"EXLA.Lib.iota/3","type":"function"},{"doc":"Returns a maximum value scalar operator for the given type. Maximum values are defined in Nx.Type.max_finite_binary/1 .","ref":"EXLA.Lib.html#max_finite/2","title":"EXLA.Lib.max_finite/2","type":"function"},{"doc":"Returns a minimum value scalar operator for the given type. Minimum values are defined in Nx.Type.min_finite_binary/1 .","ref":"EXLA.Lib.html#min_finite/2","title":"EXLA.Lib.min_finite/2","type":"function"},{"doc":"Element-wise tangent function.","ref":"EXLA.Lib.html#tan/1","title":"EXLA.Lib.tan/1","type":"function"},{"doc":"Wrapper around XLA's ops.","ref":"EXLA.Op.html","title":"EXLA.Op","type":"module"},{"doc":"Unary abs.","ref":"EXLA.Op.html#abs/1","title":"EXLA.Op.abs/1","type":"function"},{"doc":"Unary acos.","ref":"EXLA.Op.html#acos/1","title":"EXLA.Op.acos/1","type":"function"},{"doc":"Unary acosh.","ref":"EXLA.Op.html#acosh/1","title":"EXLA.Op.acosh/1","type":"function"},{"doc":"Element-wise add with broadcasting.","ref":"EXLA.Op.html#add/3","title":"EXLA.Op.add/3","type":"function"},{"doc":"Unary asin.","ref":"EXLA.Op.html#asin/1","title":"EXLA.Op.asin/1","type":"function"},{"doc":"Unary asinh.","ref":"EXLA.Op.html#asinh/1","title":"EXLA.Op.asinh/1","type":"function"},{"doc":"Element-wise atan2 with broadcasting.","ref":"EXLA.Op.html#atan2/3","title":"EXLA.Op.atan2/3","type":"function"},{"doc":"Unary atan.","ref":"EXLA.Op.html#atan/1","title":"EXLA.Op.atan/1","type":"function"},{"doc":"Unary atanh.","ref":"EXLA.Op.html#atanh/1","title":"EXLA.Op.atanh/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#bitcast_convert_type/2","title":"EXLA.Op.bitcast_convert_type/2","type":"function"},{"doc":"Element-wise bitwise_and with broadcasting.","ref":"EXLA.Op.html#bitwise_and/3","title":"EXLA.Op.bitwise_and/3","type":"function"},{"doc":"Unary bitwise_not.","ref":"EXLA.Op.html#bitwise_not/1","title":"EXLA.Op.bitwise_not/1","type":"function"},{"doc":"Element-wise bitwise_or with broadcasting.","ref":"EXLA.Op.html#bitwise_or/3","title":"EXLA.Op.bitwise_or/3","type":"function"},{"doc":"Element-wise bitwise_xor with broadcasting.","ref":"EXLA.Op.html#bitwise_xor/3","title":"EXLA.Op.bitwise_xor/3","type":"function"},{"doc":"Broadcasts the tensor to shape .","ref":"EXLA.Op.html#broadcast_in_dim/3","title":"EXLA.Op.broadcast_in_dim/3","type":"function"},{"doc":"Unary cbrt.","ref":"EXLA.Op.html#cbrt/1","title":"EXLA.Op.cbrt/1","type":"function"},{"doc":"Unary ceil.","ref":"EXLA.Op.html#ceil/1","title":"EXLA.Op.ceil/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#cholesky/1","title":"EXLA.Op.cholesky/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#clamp/3","title":"EXLA.Op.clamp/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#concatenate/2","title":"EXLA.Op.concatenate/2","type":"function"},{"doc":"","ref":"EXLA.Op.html#conditional/3","title":"EXLA.Op.conditional/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#conditional/5","title":"EXLA.Op.conditional/5","type":"function"},{"doc":"","ref":"EXLA.Op.html#conjugate/1","title":"EXLA.Op.conjugate/1","type":"function"},{"doc":"Creates a n-dimensional constant from binary data with shape .","ref":"EXLA.Op.html#constant_from_binary/3","title":"EXLA.Op.constant_from_binary/3","type":"function"},{"doc":"Creates a numeric constant.","ref":"EXLA.Op.html#constant_r0/3","title":"EXLA.Op.constant_r0/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#conv_general_dilated/10","title":"EXLA.Op.conv_general_dilated/10","type":"function"},{"doc":"","ref":"EXLA.Op.html#convert_element_type/2","title":"EXLA.Op.convert_element_type/2","type":"function"},{"doc":"Unary cos.","ref":"EXLA.Op.html#cos/1","title":"EXLA.Op.cos/1","type":"function"},{"doc":"Unary cosh.","ref":"EXLA.Op.html#cosh/1","title":"EXLA.Op.cosh/1","type":"function"},{"doc":"Unary count_leading_zeros.","ref":"EXLA.Op.html#count_leading_zeros/1","title":"EXLA.Op.count_leading_zeros/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#create_token/1","title":"EXLA.Op.create_token/1","type":"function"},{"doc":"Element-wise divide with broadcasting.","ref":"EXLA.Op.html#divide/3","title":"EXLA.Op.divide/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#dot/3","title":"EXLA.Op.dot/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#dot_general/4","title":"EXLA.Op.dot_general/4","type":"function"},{"doc":"","ref":"EXLA.Op.html#dynamic_slice/3","title":"EXLA.Op.dynamic_slice/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#dynamic_update_slice/3","title":"EXLA.Op.dynamic_update_slice/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#eigh/2","title":"EXLA.Op.eigh/2","type":"function"},{"doc":"Element-wise equal with broadcasting.","ref":"EXLA.Op.html#equal/3","title":"EXLA.Op.equal/3","type":"function"},{"doc":"Unary erf.","ref":"EXLA.Op.html#erf/1","title":"EXLA.Op.erf/1","type":"function"},{"doc":"Unary erf_inv.","ref":"EXLA.Op.html#erf_inv/1","title":"EXLA.Op.erf_inv/1","type":"function"},{"doc":"Unary erfc.","ref":"EXLA.Op.html#erfc/1","title":"EXLA.Op.erfc/1","type":"function"},{"doc":"Unary exp.","ref":"EXLA.Op.html#exp/1","title":"EXLA.Op.exp/1","type":"function"},{"doc":"Unary expm1.","ref":"EXLA.Op.html#expm1/1","title":"EXLA.Op.expm1/1","type":"function"},{"doc":"Unary floor.","ref":"EXLA.Op.html#floor/1","title":"EXLA.Op.floor/1","type":"function"},{"doc":"The XLA gather operation stitches together several slices of an input array. Note that this operation is extremely generic and far from intuitive for regular usage. However, it can be used to implement many specific operations that have to do with combining multiple tensor slices. Parameteres The XLA docs are rather cryptic unless already understood, so here's an attempt of a more intuitive description. index_vector_dim Determines which dimension contains index vectors. In most cases we want to set this to the last dimension. given start_indices = [ [ 0 , 1 ] , [ 1 , 1 ] ] and given index_vector_dim = 1 then index vectors are [ 0 , 1 ] and [ 1 , 1 ] Note that we can set this to last_dimension + 1 , in which case start_indices are implicitly reshaped to have a trailing dimension of 1. given start_indices = [ [ 0 , 1 ] , [ 1 , 1 ] ] and given index_vector_dim = 2 then start_indices &lt;- [ [ [ 0 ] , [ 1 ] ] , [ [ 1 ] , [ 1 ] ] ] index vectors are [ 0 ] , [ 1 ] , [ 1 ] , [ 1 ] start_index_map Note: though given as a list, it can be treated as a map of list_idx -&gt; value . An index vector may have less elements than the operand tensor shape. For example: given operand = [ [ 1 , 2 ] , [ 3 , 4 ] ] start_indices = [ [ 1 ] , [ 0 ] ] index_vector_dim = 1 As described above, in this case index vectors are [1] , [0] and they have length 1. However, the operand has rank 2, so we need vectors of the form [_, _] to point to a specific element in the operand. The start_index_map determines where indices go into this template: and given start_index_map = [ 0 ] # effectively %{0 =&gt; 0} then actual index vectors are [ 1 , _ ] and [ 0 , _ ] and given start_index_map = [ 1 ] # effectively %{0 =&gt; 1} then actual index vectors are [ _ , 1 ] and [ _ , 0 ] Finally, the missing elements ( _ ) are assumed to be 0. Complete examples: given operand = [ [ 1 , 2 ] , [ 3 , 4 ] ] start_indices = [ [ 0 ] , [ 1 ] ] index_vector_dim = 1 and given start_index_map = [ 1 ] # effectively %{0 =&gt; 1} then actual index vectors are [ 0 , 0 ] , [ 0 , 1 ] ( leading 0 is inserted ) given operand = [ [ 1 , 2 ] , [ 3 , 4 ] ] start_indices = [ [ 0 , 1 ] , [ 1 , 1 ] ] index_vector_dim = 1 and given start_index_map = [ 0 , 1 ] # effectively %{0 =&gt; 0, 1 =&gt; 1} then actual index vectors are [ 0 , 1 ] , [ 1 , 1 ] ( as expected ) given operand = [ [ 1 , 2 ] , [ 3 , 4 ] ] start_indices = [ [ 0 , 1 ] , [ 1 , 1 ] ] index_vector_dim = 1 and given start_index_map = [ 1 , 0 ] # effectively %{0 =&gt; 1, 1 =&gt; 0} then actual index vectors are [ 1 , 0 ] , [ 1 , 1 ] ( see how the first vector is reversed ) slice_sizes For every starting point (as described above) we take a slice given by slice_sizes . Naturally, slice_sizes must have the same length as operand rank, so that we have one size per dimension. given operand = [ [ 1 , 2 ] , [ 3 , 4 ] ] actual index vector [ 1 , 0 ] and given slice_sizes = [ 1 , 2 ] then slice for actual index vector is [ [ 3 , 4 ] ] collapsed_slice_dims A list of dimensions that are collapsed (effectively removed) in the slice shape. Only dimensions of size 1 can be collapsed. given slice is [ [ 3 , 4 ] ] # shape: [1][2] and given collapsed_slice_dims = [ 0 ] then actual slice is [ 3 , 4 ] # shape [2] offset_dims A list of dimensions in the output tensor corresponding to the non-collapsed dimensions in slice tensors. In other words, these dimensions are used for indexing elements of the slice tensors. given operand = [ [ 1 , 2 ] , [ 3 , 4 ] ] start_indices = [ [ 1 , 0 ] , [ 0 , 0 ] , [ 1 , 0 ] ] index_vector_dim = 1 start_index_map = [ 1 , 2 ] # effectively %{0 =&gt; 0, 1 =&gt; 1} collapsed_slice_dims = [ 0 ] and given offset_dims = [ 1 ] then result is [ [ 3 , 4 ] , [ 1 , 2 ] , [ 3 , 4 ] ] In the above example the collapsed slices are [3, 4] , [1, 2] , [3, 4] and have rank 1. Using offset_dims we specify that the first dimension in each slice corresponds to the second dimension in the output tensor. If we use the first output dimension instead, we get: and given offset_dims = [ 0 ] then result is [ [ 3 , 1 , 3 ] , [ 4 , 2 , 4 ] ] Docs More formal specification can be found in the XLA Gather docs .","ref":"EXLA.Op.html#gather/7","title":"EXLA.Op.gather/7","type":"function"},{"doc":"Gets the shape of an operator.","ref":"EXLA.Op.html#get_shape/1","title":"EXLA.Op.get_shape/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#get_tuple_element/2","title":"EXLA.Op.get_tuple_element/2","type":"function"},{"doc":"Element-wise greater with broadcasting.","ref":"EXLA.Op.html#greater/3","title":"EXLA.Op.greater/3","type":"function"},{"doc":"Element-wise greater_equal with broadcasting.","ref":"EXLA.Op.html#greater_equal/3","title":"EXLA.Op.greater_equal/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#imag/1","title":"EXLA.Op.imag/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#infeed/2","title":"EXLA.Op.infeed/2","type":"function"},{"doc":"Creates iota tensor.","ref":"EXLA.Op.html#iota/3","title":"EXLA.Op.iota/3","type":"function"},{"doc":"Element-wise left_shift with broadcasting.","ref":"EXLA.Op.html#left_shift/3","title":"EXLA.Op.left_shift/3","type":"function"},{"doc":"Element-wise less with broadcasting.","ref":"EXLA.Op.html#less/3","title":"EXLA.Op.less/3","type":"function"},{"doc":"Element-wise less_equal with broadcasting.","ref":"EXLA.Op.html#less_equal/3","title":"EXLA.Op.less_equal/3","type":"function"},{"doc":"Unary log1p.","ref":"EXLA.Op.html#log1p/1","title":"EXLA.Op.log1p/1","type":"function"},{"doc":"Unary log.","ref":"EXLA.Op.html#log/1","title":"EXLA.Op.log/1","type":"function"},{"doc":"Unary logistic.","ref":"EXLA.Op.html#logistic/1","title":"EXLA.Op.logistic/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#lu/1","title":"EXLA.Op.lu/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#map/3","title":"EXLA.Op.map/3","type":"function"},{"doc":"Element-wise max with broadcasting.","ref":"EXLA.Op.html#max/3","title":"EXLA.Op.max/3","type":"function"},{"doc":"Element-wise min with broadcasting.","ref":"EXLA.Op.html#min/3","title":"EXLA.Op.min/3","type":"function"},{"doc":"Element-wise multiply with broadcasting.","ref":"EXLA.Op.html#multiply/3","title":"EXLA.Op.multiply/3","type":"function"},{"doc":"Unary negate.","ref":"EXLA.Op.html#negate/1","title":"EXLA.Op.negate/1","type":"function"},{"doc":"Element-wise not_equal with broadcasting.","ref":"EXLA.Op.html#not_equal/3","title":"EXLA.Op.not_equal/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#outfeed/2","title":"EXLA.Op.outfeed/2","type":"function"},{"doc":"Pads the tensor with value and padding config.","ref":"EXLA.Op.html#pad/3","title":"EXLA.Op.pad/3","type":"function"},{"doc":"Specifies a parameter at position i with shape and name .","ref":"EXLA.Op.html#parameter/4","title":"EXLA.Op.parameter/4","type":"function"},{"doc":"Unary population_count.","ref":"EXLA.Op.html#population_count/1","title":"EXLA.Op.population_count/1","type":"function"},{"doc":"Element-wise power with broadcasting.","ref":"EXLA.Op.html#power/3","title":"EXLA.Op.power/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#qr/2","title":"EXLA.Op.qr/2","type":"function"},{"doc":"","ref":"EXLA.Op.html#real/1","title":"EXLA.Op.real/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#reduce/4","title":"EXLA.Op.reduce/4","type":"function"},{"doc":"Element-wise remainder with broadcasting.","ref":"EXLA.Op.html#remainder/3","title":"EXLA.Op.remainder/3","type":"function"},{"doc":"Reshapes the tensor to shape .","ref":"EXLA.Op.html#reshape/2","title":"EXLA.Op.reshape/2","type":"function"},{"doc":"","ref":"EXLA.Op.html#reverse/2","title":"EXLA.Op.reverse/2","type":"function"},{"doc":"Element-wise right_shift_arithmetic with broadcasting.","ref":"EXLA.Op.html#right_shift_arithmetic/3","title":"EXLA.Op.right_shift_arithmetic/3","type":"function"},{"doc":"Element-wise right_shift_logical with broadcasting.","ref":"EXLA.Op.html#right_shift_logical/3","title":"EXLA.Op.right_shift_logical/3","type":"function"},{"doc":"Creates tensor with normal distribution.","ref":"EXLA.Op.html#rng_normal/3","title":"EXLA.Op.rng_normal/3","type":"function"},{"doc":"Creates tensor with uniform distribution.","ref":"EXLA.Op.html#rng_uniform/3","title":"EXLA.Op.rng_uniform/3","type":"function"},{"doc":"Unary round.","ref":"EXLA.Op.html#round/1","title":"EXLA.Op.round/1","type":"function"},{"doc":"Unary rsqrt.","ref":"EXLA.Op.html#rsqrt/1","title":"EXLA.Op.rsqrt/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#scatter/8","title":"EXLA.Op.scatter/8","type":"function"},{"doc":"","ref":"EXLA.Op.html#select/3","title":"EXLA.Op.select/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#select_and_scatter/8","title":"EXLA.Op.select_and_scatter/8","type":"function"},{"doc":"Unary sign.","ref":"EXLA.Op.html#sign/1","title":"EXLA.Op.sign/1","type":"function"},{"doc":"Unary sin.","ref":"EXLA.Op.html#sin/1","title":"EXLA.Op.sin/1","type":"function"},{"doc":"Unary sinh.","ref":"EXLA.Op.html#sinh/1","title":"EXLA.Op.sinh/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#slice/4","title":"EXLA.Op.slice/4","type":"function"},{"doc":"","ref":"EXLA.Op.html#sort/3","title":"EXLA.Op.sort/3","type":"function"},{"doc":"Unary sqrt.","ref":"EXLA.Op.html#sqrt/1","title":"EXLA.Op.sqrt/1","type":"function"},{"doc":"Element-wise subtract with broadcasting.","ref":"EXLA.Op.html#subtract/3","title":"EXLA.Op.subtract/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#svd/2","title":"EXLA.Op.svd/2","type":"function"},{"doc":"Unary tanh.","ref":"EXLA.Op.html#tanh/1","title":"EXLA.Op.tanh/1","type":"function"},{"doc":"","ref":"EXLA.Op.html#transpose/2","title":"EXLA.Op.transpose/2","type":"function"},{"doc":"","ref":"EXLA.Op.html#triangular_solve/6","title":"EXLA.Op.triangular_solve/6","type":"function"},{"doc":"Builds a tuple with the given elements.","ref":"EXLA.Op.html#tuple/2","title":"EXLA.Op.tuple/2","type":"function"},{"doc":"","ref":"EXLA.Op.html#variadic_reduce/5","title":"EXLA.Op.variadic_reduce/5","type":"function"},{"doc":"","ref":"EXLA.Op.html#variadic_sort/4","title":"EXLA.Op.variadic_sort/4","type":"function"},{"doc":"","ref":"EXLA.Op.html#while/3","title":"EXLA.Op.while/3","type":"function"},{"doc":"","ref":"EXLA.Op.html#window_reduce/7","title":"EXLA.Op.window_reduce/7","type":"function"},{"doc":"Wrapper around XLA's shape.","ref":"EXLA.Shape.html","title":"EXLA.Shape","type":"module"},{"doc":"Returns the shape size in bytes.","ref":"EXLA.Shape.html#byte_size/1","title":"EXLA.Shape.byte_size/1","type":"function"},{"doc":"Converts a charlist type into Nx' tuple format.","ref":"EXLA.Shape.html#charlist_to_dtype/1","title":"EXLA.Shape.charlist_to_dtype/1","type":"function"},{"doc":"Converts Nx's tuple format into charlist.","ref":"EXLA.Shape.html#dtype_to_charlist/1","title":"EXLA.Shape.dtype_to_charlist/1","type":"function"},{"doc":"Creates a shape with the given type-size tuple and dimensions.","ref":"EXLA.Shape.html#make_shape/2","title":"EXLA.Shape.make_shape/2","type":"function"},{"doc":"Creates a token shape.","ref":"EXLA.Shape.html#make_token_shape/0","title":"EXLA.Shape.make_token_shape/0","type":"function"},{"doc":"Creates a tuple shape with the given shapes.","ref":"EXLA.Shape.html#make_tuple_shape/1","title":"EXLA.Shape.make_tuple_shape/1","type":"function"},{"doc":"","ref":"rotating-image.html","title":"Rotating Image","type":"extras"},{"doc":"Mix . install ( [ { :req , &quot;~&gt; 0.2.0&quot; } , { :kino , &quot;~&gt; 0.5.1&quot; } , { :exla , &quot;~&gt; 0.2&quot; } , { :evision , &quot;~&gt; 0.1.0-dev&quot; , github : &quot;cocoa-xu/evision&quot; , branch : &quot;main&quot; } ] ) # To use TPU/CUDA/ROCM, you must also set the XLA_TARGET env var EXLA . set_as_nx_default ( [ :tpu , :cuda , :rocm , :host ] )","ref":"rotating-image.html#setup","title":"Rotating Image - Setup","type":"extras"},{"doc":"Preprocessing a binary to the tensor representation and calculating shapes of the rotated image. defmodule Preprocess do def image_to_tensor ( file_or_url ) do case URI . parse ( file_or_url ) do % URI { scheme : nil } -&gt; { :ok , mat } = OpenCV . imread ( file_or_url ) OpenCV.Nx . to_nx ( mat ) _ -&gt; %{ body : binary_image } = Req . get! ( file_or_url ) { :ok , mat } = OpenCV . imdecode ( binary_image , OpenCV . cv_IMREAD_COLOR ( ) ) OpenCV.Nx . to_nx ( mat ) end end # there is no pi const in Nx yet. @pi 3.14159 defp calculate_new_side_length ( side1 , side2 , angle , radians? ) do angle = if radians? == 0 , do : angle |&gt; Nx . divide ( 180 ) |&gt; Nx . multiply ( @pi ) , else : angle sine = Nx . sin ( angle ) cosine = Nx . cos ( angle ) side1 |&gt; Nx . multiply ( cosine ) |&gt; Nx . abs ( ) |&gt; Nx . add ( Nx . abs ( Nx . multiply ( side2 , sine ) ) ) |&gt; Nx . add ( 1 ) |&gt; Nx . to_number ( ) |&gt; round ( ) end def get_new_height_and_width ( height , width , angle , radians? \\\\ 1 ) do new_height = calculate_new_side_length ( height , width , angle , radians? ) new_width = calculate_new_side_length ( width , height , angle , radians? ) { new_height , new_width } end end","ref":"rotating-image.html#preprocessing","title":"Rotating Image - Preprocessing","type":"extras"},{"doc":"defmodule Rotate do import Nx.Defn @num_channels 3 # using three shears to avoid aliasing # https://datagenetics.com/blog/august32013/index.html defnp calculate_new_positions ( { y_image , x_image , _ } , orig_ctr_height , orig_ctr_width , new_ctr_height , new_ctr_width , angle ) do k = Nx . iota ( { y_image * x_image } ) i = Nx . as_type ( Nx . divide ( k , x_image ) , { :u , 32 } ) j = rem ( k , x_image ) y = orig_ctr_height |&gt; Nx . add ( 1 ) |&gt; Nx . add ( i ) |&gt; then ( &amp; Nx . subtract ( y_image , &amp;1 ) ) x = orig_ctr_width |&gt; Nx . add ( 1 ) |&gt; Nx . add ( j ) |&gt; then ( &amp; Nx . subtract ( x_image , &amp;1 ) ) tangent = angle |&gt; Nx . divide ( 2 ) |&gt; Nx . tan ( ) new_x = y |&gt; Nx . multiply ( tangent ) |&gt; then ( &amp; Nx . subtract ( x , &amp;1 ) ) |&gt; Nx . round ( ) new_y = angle |&gt; Nx . sin ( ) |&gt; Nx . multiply ( new_x ) |&gt; Nx . add ( y ) |&gt; Nx . round ( ) new_x = new_y |&gt; Nx . multiply ( tangent ) |&gt; then ( &amp; Nx . subtract ( new_x , &amp;1 ) ) |&gt; Nx . round ( ) new_y = Nx . subtract ( new_ctr_height , new_y ) new_x = Nx . subtract ( new_ctr_width , new_x ) new_y = Nx . as_type ( Nx . round ( new_y ) , { :u , 32 } ) new_x = Nx . as_type ( Nx . round ( new_x ) , { :u , 32 } ) Nx . stack ( [ new_y , new_x ] , axis : 1 ) end defnp calculate_ctr ( coordinate ) do coordinate |&gt; Nx . add ( 1 ) |&gt; Nx . divide ( 2 ) |&gt; Nx . subtract ( 1 ) |&gt; Nx . round ( ) end defnp preprocess_position ( pos , n ) do Nx . concatenate ( [ Nx . new_axis ( pos , 1 ) |&gt; Nx . tile ( [ 1 , @num_channels , 1 ] ) , Nx . iota ( { n , @num_channels , 1 } , axis : 1 ) ] , axis : 2 ) |&gt; Nx . reshape ( { n * @num_channels , 3 } ) end # there is no pi const in Nx yet. @pi 3.14159 # new_image is a tensor with dims of rotated image filled with zeros defn rotate_image_by_angle ( image , angle , new_image , radians? \\\\ 1 ) do # if radians? is set to false than assuming that angle is given in degrees angle = if radians? == 0 , do : angle |&gt; Nx . divide ( 180 ) |&gt; Nx . multiply ( @pi ) , else : angle { height , width , _ } = image . shape { new_height , new_width , _ } = new_image . shape orig_ctr_height = calculate_ctr ( height ) orig_ctr_width = calculate_ctr ( width ) new_ctr_height = calculate_ctr ( new_height ) new_ctr_width = calculate_ctr ( new_width ) pos = calculate_new_positions ( image . shape , orig_ctr_height , orig_ctr_width , new_ctr_height , new_ctr_width , angle ) { n , 2 } = Nx . shape ( pos ) preprocessed_pos = preprocess_position ( pos , n ) image = image |&gt; Nx . reshape ( { n * @num_channels } ) Nx . indexed_add ( new_image , preprocessed_pos , image ) end end","ref":"rotating-image.html#rotating-image","title":"Rotating Image - Rotating image","type":"extras"},{"doc":"num_channels = 3 dir = File . cwd! ( ) default_input = &quot;https://www.researchgate.net/publication/259521525/figure/fig8/AS:268024322195479@1440913386352/Original-standard-test-image-of-Peppers.png&quot; input = Kino.Input . text ( &quot;Image to rotate&quot; , default : default_input ) Kino . render ( input ) path_to_file = Kino.Input . read ( input ) # set options radians? = 0 angle = 70 # rotate image by 70 degrees to the right image = Preprocess . image_to_tensor ( path_to_file ) { height , width , _ } = image . shape { new_height , new_width } = Preprocess . get_new_height_and_width ( height , width , angle , radians? ) new_image = Nx . broadcast ( Nx . tensor ( [ 0 ] , type : { :u , 8 } ) , { new_height , new_width , num_channels } ) rotated_image = Rotate . rotate_image_by_angle ( image , angle , new_image , radians? ) { :ok , rotated_image_mat } = OpenCV.Nx . to_mat ( rotated_image ) { :ok , content } = OpenCV . imencode ( &quot;.png&quot; , rotated_image_mat ) content = :binary . list_to_bin ( content ) # display the rotated image Kino.Image . new ( content , &quot;image/png&quot; )","ref":"rotating-image.html#example-of-usage","title":"Rotating Image - Example of usage","type":"extras"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"First release.","ref":"changelog.html#v0-1-0-2022-04-28","title":"Changelog - v0.1.0 (2022-04-28)","type":"extras"}]